from typing import BinaryIO
from dataclasses import dataclass
import io


@dataclass
class Fragment:
    offset: int
    size: int
    data: bytes | bytearray | memoryview | None = None

    def set_data(self, data: bytes | bytearray | memoryview):
        if self.size != len(data):
            raise ValueError('Size mismatch.')
        self.data = data


class BinaryBuilder:
    _fragments: list[Fragment]
    _last_offset: int

    def __init__(self, base: int = 0):
        self._fragments = []
        self._last_offset = base

    def append(self, size: int) -> Fragment:
        result = Fragment(self._last_offset, size, None)
        self._fragments.append(result)
        self._last_offset += size
        return result

    def concat(self) -> bytes:
        result = io.BytesIO()
        for fragment in self._fragments:
            result.write(fragment.data)
        return result.getvalue()

    def sizeof(self):
        return self._last_offset


def simple_checksum(input_file: BinaryIO, size: int | None = None) -> int:
    buf = bytearray(1024)
    buf_mv = memoryview(buf)
    checksum = 0

    if size is None:
        old_pos = input_file.tell()
        bytes_left = input_file.seek(0, io.SEEK_END) - old_pos
        input_file.seek(old_pos, io.SEEK_SET)
    else:
        bytes_left = size

    while True:
        if bytes_left == 0:
            break

        actual = input_file.readinto(buf)

        if actual == 0:
            break
        elif actual > bytes_left:
            actual = bytes_left

        checksum += sum(buf_mv[:actual])
        if bytes_left is not None:
            bytes_left -= actual

    return checksum & 0xffff



from typing import cast
from collections.abc import Generator, Sequence

from dataclasses import dataclass
from pathlib import Path

import usb.core


@dataclass
class DfuType:
    name: str
    description: str
    vid: int
    pid: int
    manufacturer: str
    product: str

    def emit_udev_rule(self):
        return (
            f'# {self.description}\n'
            f'SUBSYSTEM="usb", ATTR{{idVendor}}=="{self.vid:04x}", ATTR{{idProduct}}=="{self.pid:04x}", '
            f'ATTR{{manufacturer}}=="{self.manufacturer}", ATTR{{product}}=="{self.product}", TAG+="uaccess"'
        )


KNOWN_DEVICE_TYPES: Sequence[DfuType] = (
    DfuType(
        name='s3c',
        description='S3C2416 family bootloader',
        vid=0x054b,
        pid=0x1582,
        manufacturer='SAMSUNG',
        product='S3C2443',
    ),
    DfuType(
        name='imx233',
        description='iMX233 family bootloader',
        vid=0x0419,
        pid=0x0133,
        manufacturer='Besta',
        product='UDisk',
    ),
    DfuType(
        name='w55pa71',
        description='Nuvoton W55PA71 family bootloader',
        vid=0x0417,
        pid=0x9394,
        manufacturer='Winbond      ',
        product='W55PA71      ',
    ),
    DfuType(
        name='fa32',
        description='Nuvoton FA32 family bootloader',
        vid=0x0483,
        pid=0x5720,
        manufacturer='nuvoTon',
        product='USB MSC',
    ),
)

KNOWN_DEVICE_INDEX: dict[tuple[int | None, int | None, str | None, str | None], DfuType] = {
    (dev.vid, dev.pid, dev.manufacturer, dev.product): dev for dev in KNOWN_DEVICE_TYPES
}


def generate_udev_file(path: Path) -> None:
    with path.open('w') as udev_file:
        udev_file.write('# Generated by dfutool from besta-tools\n\n')
        for dev_type in KNOWN_DEVICE_TYPES:
            udev_file.write(dev_type.emit_udev_rule())
            udev_file.write('\n')


def trim_nul_terminated(value: str | None) -> str | None:
    if value is None:
        return None
    first_zero = value.find('\x00')
    if first_zero == -1:
        return value
    return value[:first_zero]


def enumerate_device() -> Generator[tuple[DfuType, usb.core.Device]]:
    for dev in usb.core.find(find_all=True):
        try:
            device_type_key = (
                cast(int | None, dev.idVendor),
                cast(int | None, dev.idProduct),
                trim_nul_terminated(dev.manufacturer),
                trim_nul_terminated(dev.product)
            )
            if device_type_key in KNOWN_DEVICE_INDEX:
                yield KNOWN_DEVICE_INDEX[device_type_key], dev
        except ValueError:
            # No langid and probably no access to device. Skipping,
            continue



USB_INTERFACE_SUBCLASS_NONE = 0x00
"SCSI command set not reported."
USB_INTERFACE_SUBCLASS_RBC = 0x01
"Reduced Block Command (RBC)."
USB_INTERFACE_SUBCLASS_MMC5 = 0x02
"MMC-5 (ATAPI)."
USB_INTERFACE_SUBCLASS_QIC157 = 0x03
"QIC-157 (obsolete)."
USB_INTERFACE_SUBCLASS_UFI = 0x04
"UFI (floppy drive)."
USB_INTERFACE_SUBCLASS_SFF8070I = 0x05
"SFF-8070i (obsolete)."
USB_INTERFACE_SUBCLASS_SCSI = 0x06
"SCSI transparent command set."
USB_INTERFACE_SUBCLASS_LSDFS = 0x07
"Lockable Storage Devices (LSD) feature specification."
USB_INTERFACE_SUBCLASS_IEEE1667 = 0x08
"IEEE-1667."

USB_INTERFACE_PROTOCOL_CBI_WITH_COMPLETION = 0x00
"Control/Block/Interrupt (CBI) (with command completion interrupt)."
USB_INTERFACE_PROTOCOL_CBI = 0x01
"Control/Block/Interrupt (CBI) (without command completion interrupt)."
USB_INTERFACE_PROTOCOL_BBB = 0x50
"Bulk-only (BBB)."
USB_INTERFACE_PROTOCOL_UAS = 0x62
"USB Attached SCSI (UAS)."

USBMS_REQ_ADSC = 0x00
"Accept Device-Specific Command (ADSC)"
USBMS_REQ_LSD_GET_REQUESTS = 0xfc
"LSD Get Requests"
USBMS_REQ_LSD_PUT_REQUESTS = 0xfd
"LSD Put Requests"
USBMS_REQ_BBB_GET_MAX_LUN = 0xfe
"Bulk-only Get Max LUN"
USBMS_REQ_BBB_RESET = 0xff
"Bulk-only Reset"

USBMS_CBW_MAGIC = 0x43425355
USBMS_CSW_MAGIC = 0x53425355

USBMS_CBW_DIR_MASK = 0x80
USBMS_CBW_DIR_IN = 0x80
USBMS_CBW_DIR_OUT = 0x00

USBMS_CSW_STATUS_GOOD = 0x00
USBMS_CSW_STATUS_BAD = 0x01
USBMS_CSW_STATUS_PHASE_ERROR = 0x02

SCSI_CMD_TEST_UNIT_READY = 0x00
SCSI_CMD_REQUEST_SENSE = 0x03
SCSI_CMD_INQUIRY = 0x12
SCSI_CMD_MODE_SENSE6 = 0x1a
SCSI_CMD_START_STOP_UNIT = 0x1b
SCSI_CMD_PREVENT_ALLOW_MEDIUM_REMOVAL = 0x1e
SCSI_CMD_READ_FORMAT_CAPACITY = 0x23
SCSI_CMD_READ_CAPACITY10 = 0x25
SCSI_CMD_READ10 = 0x28
SCSI_CMD_WRITE10 = 0x2a
SCSI_CMD_VERIFY10 = 0x2f
SCSI_CMD_MODE_SENSE10 = 0x5a

SCSI_REQUEST_SENSE_F0_ERROR_CODE_MASK = 0x7f
SCSI_REQUEST_SENSE_F0_VALID_MASK = 0x80
SCSI_ERROR_CODE_CURRENT = 0x70
SCSI_ERROR_CODE_DEFERRED = 0x71

SCSI_REQUEST_SENSE_F2_SENSE_KEY_MASK = 0x0f
SCSI_SENSE_KEY_NO_SENSE = 0x0
SCSI_SENSE_KEY_RECOVERED_ERROR = 0x1
SCSI_SENSE_KEY_NOT_READY = 0x2
SCSI_SENSE_KEY_MEDIUM_ERROR = 0x3
SCSI_SENSE_KEY_HARDWARE_ERROR = 0x4
SCSI_SENSE_KEY_ILLEGAL_REQUEST = 0x5
SCSI_SENSE_KEY_UNIT_ATTENTION = 0x6
SCSI_SENSE_KEY_DATA_PROTECT = 0x7
SCSI_SENSE_KEY_BLANK_CHECK = 0x8
SCSI_SENSE_KEY_VENDOR_SPECIFIC = 0x9
SCSI_SENSE_KEY_COPY_ABORTED = 0xa
SCSI_SENSE_KEY_ABORTED_COMMAND = 0xb
SCSI_SENSE_KEY_VOLUME_OVERFLOW = 0xd
SCSI_SENSE_KEY_MISCOMPARE = 0xe

SCSI_ASC_NONE = 0x00
SCSI_ASC_LUN_NOT_READY = 0x0400
SCSI_ASC_LUN_STARTING = SCSI_ASC_LUN_NOT_READY | 0x01
SCSI_ASC_INVALID_FIELD_IN_CDB = 0x2400
SCSI_ASC_LUN_NOT_SUPPORTED = 0x2500

SCSI_INQUIRY_HEADER_F1_RMB = 0x80
SCSI_INQUIRY_HEADER_F3_AERC = 0x80
SCSI_INQUIRY_HEADER_F3_NORMACA = 0x20
SCSI_INQUIRY_HEADER_F3_HISUP = 0x10
SCSI_INQUIRY_HEADER_F3_FORMAT_MASK = 0x0f
SCSI_INQUIRY_HEADER_FORMAT_SPC2 = 0x02

SCSI_INQUIRY_HEADER_VERSION_NONE = 0x0
SCSI_INQUIRY_HEADER_VERSION_SCSI2_ANSI = 0x2
SCSI_INQUIRY_HEADER_VERSION_SPC = 0x3
SCSI_INQUIRY_HEADER_VERSION_SPC2 = 0x4
SCSI_INQUIRY_HEADER_VERSION_SCSI2_ISO = 0x80

SCSI_INQUIRY_PERIF_QUALIFIER_LOADED = 0 << 5
SCSI_INQUIRY_PERIF_QUALIFIER_UNLOADED = 1 << 5
SCSI_INQUIRY_PERIF_QUALIFIER_UNKNOWN = 3 << 5
SCSI_INQUIRY_PERIF_TYPE_DIRECT = 0x00

SCSI_INQUIRY_FEATURES_F0_SCCS = 0x80
SCSI_INQUIRY_FEATURES_FLAG_BQUE = 0x8000
SCSI_INQUIRY_FEATURES_FLAG_ENCSERV = 0x4000
SCSI_INQUIRY_FEATURES_FLAG_MULTIP = 0x1000
SCSI_INQUIRY_FEATURES_FLAG_MCHNGR = 0x0800
SCSI_INQUIRY_FEATURES_FLAG_ADDR16 = 0x0100
SCSI_INQUIRY_FEATURES_FLAG_RELADR = 0x0080
SCSI_INQUIRY_FEATURES_FLAG_WBUS16 = 0x0020
SCSI_INQUIRY_FEATURES_FLAG_SYNC = 0x0010
SCSI_INQUIRY_FEATURES_FLAG_LINKED = 0x0008
SCSI_INQUIRY_FEATURES_FLAG_CMDQUE = 0x0002

SCSI_VERIFY_F1_BYTCHK = 0x02



from typing import cast, TYPE_CHECKING
if TYPE_CHECKING:
    from construct import Context

import dataclasses
from construct import (
    Check,
    Int16ul,
    Rebuild,
    this
)
from construct_typed import DataclassMixin, DataclassStruct, csfield


def _inv_u16_per_byte(ctx: 'Context') -> int:
    value = cast(int, ctx.checksum)
    lo = value & 0xff
    hi = (value >> 8) & 0xff
    return (((0x100 - hi) & 0xff) << 8) | ((0x100 - lo) & 0xff)


@dataclasses.dataclass
class ChecksumValue(DataclassMixin):
    checksum: int = csfield(Int16ul)
    checksum_byteinv: int = csfield(Rebuild(Int16ul, _inv_u16_per_byte))
    _integrity: None = csfield(Check(
        (
            (this.checksum & 0xff) +
            (this.checksum_byteinv & 0xff) +
            ((this.checksum >> 8) & 0xff) +
            ((this.checksum_byteinv >> 8) & 0xff)
        ) & 0xff == 0
    ))


CsChecksumValue = DataclassStruct(ChecksumValue)



from typing import BinaryIO

import tomllib

import click

from besta_tools.romtool.builder import build_embeddable_from_spec_file


@click.group()
def app():
    pass

@app.command(name='build')
@click.argument('spec-file', type=click.File('rb'))
@click.option('-o', '--output', type=click.File('wb'))
def do_build(spec_file: BinaryIO, output: BinaryIO):
    spec_dict = tomllib.load(spec_file)
    spec_built = build_embeddable_from_spec_file(spec_dict)
    output.write(spec_built)



import dataclasses
import datetime

from construct import Byte, Const, Default, Int16ul, Int32ul, this
from construct_typed import DataclassMixin, DataclassStruct, EnumBase, TEnum, csfield

from ..common.formats import CsChecksumValue, ChecksumValue


class RomType(EnumBase):
    DATA_BOOK = 0x8
    DATA_DICT = 0x9
    APPLET_GAME = 0xb
    APPLET_TOOL = 0x11
    APPLET_STUDY = 0x16


# CsRomType = TEnum(Int16ul, RomType)


class RomLocale(EnumBase):
    NONE = 0x0
    ZH_CN = 0x1
    ZH_TW = 0x2
    JA_JP = 0x4
    KO_KR = 0x8
    TH_TH = 0x10
    EN_US = 0x1000
    FORCE_UTF16 = 0x8000
    UNSET = 0xffff


# CsRomLocale = TEnum(Int16ul, RomLocale)


class MagicType(EnumBase):
    APPLET = (ord('I') | (ord('F') << 8)) & 0xffff
    DATA = (ord('M') | (ord('F') << 8)) & 0xffff


CsMagicType = TEnum(Int16ul, MagicType)


@dataclasses.dataclass
class RomSpecTimestamp(DataclassMixin):
    year: int = csfield(Int16ul)
    month: int = csfield(Int16ul)
    day: int = csfield(Int16ul)
    hour: int = csfield(Int16ul)
    minute: int = csfield(Int16ul)
    second: int = csfield(Int16ul)

    @classmethod
    def from_date(cls, date: datetime.datetime):
        return cls(
            year=date.year,
            month=date.month,
            day=date.day,
            hour=date.hour,
            minute=date.minute,
            second=date.second,
        )

    @classmethod
    def now(cls):
        now = datetime.datetime.now(tz=datetime.UTC)
        return cls.from_date(now)


CsRomSpecTimestamp = DataclassStruct(RomSpecTimestamp)


@dataclasses.dataclass
class RomSpecType(DataclassMixin):
    magic: int = csfield(CsMagicType)
    spec_size: int = csfield(Const(0x80, Int16ul))
    type_: RomType | int = csfield(Int16ul)  # TODO RomType does not work as flags.
    checksum: ChecksumValue = csfield(CsChecksumValue)
    unk_0xa: int = csfield(Default(Int16ul, 0xffff))
    default_locale: RomLocale | int = csfield(Int16ul)  # TODO RomLocale does not work as flags.
    sections_offset: int = csfield(Int16ul)
    build_timestamp: RomSpecTimestamp = csfield(CsRomSpecTimestamp)
    rom_size: int = csfield(Int32ul)
    entry_point: int = csfield(Int32ul)
    fallback_title_offset: int = csfield(Int32ul)
    copyright_offset: int = csfield(Int32ul)
    icon_offset: int = csfield(Int32ul)
    version_offset: int = csfield(Int32ul)
    data_offset: int = csfield(Int32ul)
    sdk_id_offset: int = csfield(Int32ul)
    unk_0x3c: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x40: int = csfield(Default(Int32ul, 0xffffffff))
    ext_metadata_offset: int = csfield(Int32ul)
    unk_0x48: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x4c: int = csfield(Default(Int16ul, 0xffff))
    unk_0x4e: int = csfield(Default(Int16ul, 0xffff))
    localized_title_offset: int = csfield(Int32ul)
    unk_0x54: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x58: int = csfield(Default(Int32ul, 0xffffffff))
    subtype: int = csfield(Int32ul)
    unk_0x60: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x64: list[int] = csfield(Default(Int32ul[7], tuple([0xffffffff] * 7)))


CsRomSpecType = DataclassStruct(RomSpecType)


@dataclasses.dataclass
class RomFallbackTitle(DataclassMixin):
    unk_0x0: int = csfield(Default(Int16ul, 0xffff))
    locale: RomLocale | int = csfield(Int16ul)
    title_offset: int = csfield(Int32ul)
    chinese_title_offset: int = csfield(Int32ul)
    short_title_offset: int = csfield(Int32ul)
    chinese_short_title_offset: int = csfield(Int32ul)


CsRomFallbackTitle = DataclassStruct(RomFallbackTitle)


@dataclasses.dataclass
class RomExtMetadataHeader(DataclassMixin):
    unk_0x0: list[int] = csfield(Default(Int32ul[5], tuple([0xffffffff] * 5)))
    num_localized_title_entries: int = csfield(Int32ul)
    unk_0x18: list[int] = csfield(Default(Int32ul[3], tuple([0xffffffff] * 3)))


CsRomExtMetadataHeader = DataclassStruct(RomExtMetadataHeader)


@dataclasses.dataclass
class RomLocalizedTitle(DataclassMixin):
    num_entries: int = csfield(Int32ul)
    locale: RomLocale | int = csfield(Int16ul)
    unk_0x6: int = csfield(Default(Int16ul, 0x0))
    unk_0x8: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0xc: int = csfield(Default(Int32ul, 0xffffffff))


CsRomLocalizedTitle = DataclassStruct(RomLocalizedTitle)


@dataclasses.dataclass
class RomLocalizedTitleEntry(DataclassMixin):
    locale: int | RomLocale = csfield(Int16ul)
    unk_0x2: int = csfield(Default(Int16ul, 0x0))
    unk_0x4: int = csfield(Default(Int32ul, 0xffffffff))
    title_offset: int = csfield(Int32ul)
    short_title_offset: int = csfield(Int32ul)
    unk_0x10: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x14: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x18: int = csfield(Default(Int32ul, 0xffffffff))
    unk_0x1c: int = csfield(Default(Int32ul, 0xffffffff))


CsRomLocalizedTitleEntry = DataclassStruct(RomLocalizedTitleEntry)


@dataclasses.dataclass
class RomExecutableHeaderType(DataclassMixin):
    header_size: int = csfield(Const(0x80, Int16ul))
    code_size: int = csfield(Int32ul)
    unk_0x8: int = csfield(Int32ul)
    unk_0xc: int = csfield(Int32ul)
    unk_0x10: int = csfield(Int32ul)
    unk_0x14: int = csfield(Int32ul)
    unk_0x18: int = csfield(Int32ul)
    unk_0x1c: int = csfield(Int32ul)
    unk_0x20: int = csfield(Int32ul)
    unk_0x24: int = csfield(Int32ul)
    arch: int = csfield(Byte)
    unk_0x29: int = csfield(Byte)
    unk_0x2a: int = csfield(Int16ul)
    _reserved: list[int] = csfield(Int32ul[21])



#!/usr/bin/env python3
from __future__ import annotations

import re
from typing import (
    Any,
    Tuple,
    Dict,
    Sequence,
    Optional,
    BinaryIO,
    List,
    Union,
)

from collections import defaultdict

import argparse
import datetime
import enum
import io
import logging
import os
import shutil

import pefile
from elftools.elf.elffile import ELFFile
from elftools.elf import constants as elfconsts
from elftools.elf import enums as elfenums
from elftools.elf.relocation import RelocationSection

from besta_tools.common.utils import BinaryBuilder

PEFILE_WARNING_TOO_MANY_0 = re.compile(
    r"Byte 0x00 makes up [0-9]{2}\.?[0-9]+% of the file's contents. "
    r"This may indicate truncation / malformation."
)

ELF2BESTAPE_VERSION = (1, 0, 0)


logger = logging.getLogger('elf2bestape')

def parse_loglevel(level: str) -> Union[int, str]:
    try:
        return int(level)
    except ValueError:
        return level


PEStructureDefinition = Tuple[str, Sequence[str]]


def pefile_struct_calcsize(format_: PEStructureDefinition) -> int:
    pe_struct = pefile.Structure(format_)
    return pe_struct.sizeof()


def pefile_struct_from_dict(format_: PEStructureDefinition, data: Dict[str, Any], name: Optional[str] = None, file_offset: Optional[int] = None) -> pefile.Structure:
    pe_struct = pefile.Structure(format_, name, file_offset)
    pe_struct.__dict__.update(data)
    return pe_struct


ENUM_RELOC_NAME_ARM = {v: k for k, v in elfenums.ENUM_RELOC_TYPE_ARM.items()}

AAELF_RELOC_ABSOLUTE = tuple(elfenums.ENUM_RELOC_TYPE_ARM[n] for n in (
    'R_ARM_ABS32',
    'R_ARM_TARGET1',
))

R_ARM_PREL31 = elfenums.ENUM_RELOC_TYPE_ARM['R_ARM_PREL31']

AAELF_RELOC_RELATIVE = tuple(elfenums.ENUM_RELOC_TYPE_ARM[n] for n in (
    'R_ARM_REL32',
    'R_ARM_TARGET2',
))

AAELF_RELOC_IGNORE = tuple(elfenums.ENUM_RELOC_TYPE_ARM[n] for n in (
    'R_ARM_CALL', # Relative offset will still be correct if we don't shift anything
    'R_ARM_THM_CALL', # Same as above
    'R_ARM_JUMP24', # Same as R_ARM_CALL
    'R_ARM_V4BX', # just a marker for v4T BX and safe to ignore
    'R_ARM_NONE', # Doesn't seem to be useful in an executable image
    'R_ARM_TLS_LE32', # TODO is this really the right thing to do?
    'R_ARM_TLS_IE32', # PC relative offset and should be handled by GCC codegen
))

EMPTY_DOS_HEADER = {
    'e_magic': pefile.IMAGE_DOS_SIGNATURE,
    'e_cblp': 0x0,
    'e_cp': 0x0,
    'e_crlc': 0x0,
    'e_cparhdr': 0x0,
    'e_minalloc': 0x0,
    'e_maxalloc': 0x0,
    'e_ss': 0x0,
    'e_sp': 0x0,
    'e_csum': 0x0,
    'e_ip': 0x0,
    'e_cs': 0x0,
    'e_lfarlc': 0x0,
    'e_ovno': 0x0,
    'e_res': b'\x00\x00\x00\x00\x00\x00\x00\x00',
    'e_oemid': 0x0,
    'e_oeminfo': 0x0,
    'e_res2': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
    'e_lfanew': 0x40,
}

NT_HEADER = { 'Signature': pefile.IMAGE_NT_SIGNATURE }

EMPTY_FILE_HEADER = {
    'Machine': pefile.MACHINE_TYPE['IMAGE_FILE_MACHINE_ARM'], # see pefile.MACHINE_TYPE
    'NumberOfSections': 0,
    'TimeDateStamp': 0,
    'PointerToSymbolTable': 0x0,
    'NumberOfSymbols': 0x0,
    'SizeOfOptionalHeader': 0, # To be filled later.
    'Characteristics': 0, # see pefile.IMAGE_CHARACTERISTICS
}

EMPTY_OPTIONAL_HEADER = {
    "Magic": pefile.OPTIONAL_HEADER_MAGIC_PE,
    "MajorLinkerVersion": ELF2BESTAPE_VERSION[0],
    "MinorLinkerVersion": ELF2BESTAPE_VERSION[1],
    "SizeOfCode": 0,
    "SizeOfInitializedData": 0,
    "SizeOfUninitializedData": 0,
    "AddressOfEntryPoint": 0,
    "BaseOfCode": 0,
    "BaseOfData": 0,
    "ImageBase": 0,
    "SectionAlignment": 0x1000,
    "FileAlignment": 0x200,
    "MajorOperatingSystemVersion": 4, # NT 4.0 (?)
    "MinorOperatingSystemVersion": 0,
    "MajorImageVersion": 1,
    "MinorImageVersion": 0,
    "MajorSubsystemVersion": 4,
    "MinorSubsystemVersion": 0,
    "Reserved1": 0,
    "SizeOfImage": 0,
    "SizeOfHeaders": 0, # Calculate from adding all header sizes
    "CheckSum": 0,
    "Subsystem": pefile.SUBSYSTEM_TYPE['IMAGE_SUBSYSTEM_WINDOWS_CE_GUI'], # pefile.SUBSYSTEM_TYPE
    "DllCharacteristics": 0, # pefile.DLL_CHARACTERISTICS
    "SizeOfStackReserve": 1 * 1024 * 1024,
    "SizeOfStackCommit": 4096,
    "SizeOfHeapReserve": 1 * 1024 * 1024,
    "SizeOfHeapCommit": 4096,
    "LoaderFlags": 0,
    "NumberOfRvaAndSizes": pefile.IMAGE_NUMBEROF_DIRECTORY_ENTRIES,
}

EMPTY_SECTION_HEADER = {
    "Name": b'',
    "Misc_VirtualSize": 0,
    "VirtualAddress": 0,
    "SizeOfRawData": 0,
    "PointerToRawData": 0,
    "PointerToRelocations": 0,
    "PointerToLinenumbers": 0,
    "NumberOfRelocations": 0,
    "NumberOfLinenumbers": 0,
    "Characteristics": 0, # pefile.SECTION_CHARACTERISTICS
}

EMPTY_IMAGE_RESOURCE_DIRECTORY = {
    'Characteristics': 0,
    'TimeDateStamp': 0,
    'MajorVersion': 0,
    'MinorVersion': 0,
    'NumberOfNamedEntries': 0,
    'NumberOfIdEntries': 0,
}

EMPTY_IMAGE_RESOURCE_DIRECTORY_ENTRY = {
    'Name': 0x0,
    'OffsetToData': 0x0,
}

EMPTY_IMAGE_RESOURCE_DATA_ENTRY = {
    'OffsetToData': 0x0,
    'Size': 0x0,
    'CodePage': 0x0,
    'Reserved': 0x0,
}

IMAGE_RESOURCE_DIRECTORY_SIZE = pefile_struct_calcsize(pefile.PE.__IMAGE_RESOURCE_DIRECTORY_format__)
IMAGE_RESOURCE_DIRECTORY_ENTRY_SIZE = pefile_struct_calcsize(pefile.PE.__IMAGE_RESOURCE_DIRECTORY_ENTRY_format__)
IMAGE_RESOURCE_DATA_ENTRY_SIZE = pefile_struct_calcsize(pefile.PE.__IMAGE_RESOURCE_DATA_ENTRY_format__)

RSRC_STR_BIN = 'BIN'
RSRC_STR_ROMSPC = 'ROMSPC'
RSRC_NAME_BIN = len(RSRC_STR_BIN).to_bytes(2, 'little') + RSRC_STR_BIN.encode('utf-16le')
RSRC_NAME_ROMSPC = len(RSRC_STR_ROMSPC).to_bytes(2, 'little') + RSRC_STR_ROMSPC.encode('utf-16le')
RSRC_NAME_NIL = (0).to_bytes(2, 'little')

IMAGE_RESOURCE_DIRECTORY_ENTRY_ISDIR = IMAGE_RESOURCE_DIRECTORY_ENTRY_ISNAME = 0x80000000

BESTAPE_MAX_HEADER_SIZE = 0x1000 # 1 memory page
BESTAPE_MAX_ROMSPEC_SIZE = 0x8000 # 8 memory page or 32KiB


class BestaPESection(enum.IntEnum):
    SCN_TEXT = 0
    SCN_RDATA = 1
    SCN_DATA = 2
    SCN_RELOC = 3
    TOTAL = 4


def align(pos: int, blksize: int, greedy: bool = False) -> int:
    return (pos // blksize * blksize) + (blksize if greedy or pos % blksize != 0 else 0)

def lalign(pos: int, blksize: int) -> int:
    return pos // blksize * blksize

def lpadding(pos: int, blksize: int) -> int:
    return pos - (pos // blksize * blksize)

def parse_args() -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    p = argparse.ArgumentParser(description='Generate Besta PE file from an AAELF file.')
    p.add_argument('elf', help='Input AAELF file.')
    p.add_argument('-l', '--log-level', type=parse_loglevel, default='INFO', help='Set log level.')
    p.add_argument('-o', '--output', help='Besta PE file to output (or AAELF\'s basename + .exe if not supplied).')
    p.add_argument('-r', '--romspec-file', help='Include binary ROM spec file to make Type 2 ROM file.')
    p.add_argument('--deterministic', action='store_true', default=False, help='Enable deterministic conversion. (i.e. omitting fields that may affect the hash of the binary such as build timestamp)')
    return p, p.parse_args()

def generate_padding(length: int, blksize: int, greedy: bool = False) -> bytes:
    return b'\x00' * (align(length, blksize, greedy=greedy) - length)

def get_executable_segment(elf):
    for idx, seg in enumerate(elf.iter_segments()):
        if seg['p_type'] == 'PT_LOAD' and seg['p_flags'] == (elfconsts.P_FLAGS.PF_R | elfconsts.P_FLAGS.PF_X):
            return idx
    raise RuntimeError('Cannot find executable segment.')

def convert(elf_file: BinaryIO, pe_file: io.BytesIO, romspec: Optional[bytes], args: argparse.Namespace):
    elf = ELFFile(elf_file)

    # Pass 1: Setup basic PE header fields
    optional_header_dict = EMPTY_OPTIONAL_HEADER.copy()
    text_base = 0
    data_base = 0
    text_size = 0
    data_size = 0
    bss_size = 0
    rsrc_base = 0
    reloc_base = 0
    text_data: Optional[Tuple[bytes, int, bool]] = None
    rdata_data: Optional[Tuple[bytes, int, bool]] = None
    data_data: Optional[Tuple[bytes, int, bool]] = None
    rsrc_data: Optional[Tuple[bytes, int, bool]] = None
    executable_seg = get_executable_segment(elf)
    elf_base = elf.get_segment(executable_seg)['p_vaddr']
    image_base = elf_base - BESTAPE_MAX_HEADER_SIZE

    patches: Dict[int, bytes] = {}

    # Generate section headers
    section_dicts = []
    segment_lpaddings = {}
    segment_load_file_size = 0
    for idx, seg in enumerate(elf.iter_segments()):
        sec_header_dict = EMPTY_SECTION_HEADER.copy()
        # Workaround data alignment issue in stock GCC ldscript.
        lpad = lpadding(seg['p_vaddr'] - image_base, 0x1000)
        if seg['p_type'] == 'PT_LOAD':
            if seg['p_flags'] == elfconsts.P_FLAGS.PF_R | elfconsts.P_FLAGS.PF_X:
                logger.info('Found segment that maps to .text at segment #%d', idx)
                if text_data is not None:
                    raise RuntimeError('.text section already exists.')
                sec_header_dict['Name'] = b'.text'
                sec_header_dict['Characteristics'] = (
                    pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_CODE'] |
                        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ'] |
                        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_EXECUTE']
                )
                text_base = seg['p_vaddr'] - image_base
                text_size = align(seg['p_filesz'], 0x200)
                text_data = (seg.data(), lpad, False)
            elif seg['p_flags'] == elfconsts.P_FLAGS.PF_R:
                logger.info('Found segment that maps to .rdata at segment #%d', idx)
                if rdata_data is not None:
                    raise RuntimeError('.rdata section already exists.')
                sec_header_dict['Name'] = b'.rdata'
                sec_header_dict['Characteristics'] = (
                    pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA'] |
                        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ']
                )
                data_base = seg['p_vaddr'] - image_base
                data_size += align(seg['p_filesz'], 0x200)
                rdata_data = (seg.data(), lpad, False)
            elif seg['p_flags'] == elfconsts.P_FLAGS.PF_R | elfconsts.P_FLAGS.PF_W:
                logger.info('Found segment that maps to .data at segment #%d', idx)
                if data_data is not None:
                    raise RuntimeError('.data section already exists.')
                sec_header_dict['Name'] = b'.data'
                sec_header_dict['Characteristics'] = (
                    pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA'] |
                        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ'] |
                        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_WRITE']
                )
                data_size += align(seg['p_filesz'], 0x200)
                bss_size = seg['p_memsz'] - seg['p_filesz']
                if bss_size != 0:
                    sec_header_dict['Characteristics'] |= pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_UNINITIALIZED_DATA']
                data_data = (seg.data(), lpad, False)
                # PE relocation table begins after the end of all ELF segments.
                rsrc_base = align(seg['p_vaddr'] + seg['p_memsz'] - image_base, 0x1000)
            else:
                raise RuntimeError(f'Unknown PT_LOAD segment {idx} with flag {seg["p_flags"]:#010x}')
            segment_load_file_size += seg['p_filesz']
        elif seg['p_type'] in ('PT_ARM_EXIDX', 'PT_TLS'):
            continue
        else:
            raise RuntimeError(f'Unknown segment {seg["p_type"]}')
        sec_header_dict['Misc_VirtualSize'] = sec_header_dict['Misc_PhysicalAddress'] = sec_header_dict['Misc'] = seg['p_memsz']
        sec_header_dict['VirtualAddress'] = lalign(seg['p_vaddr'] - image_base, 0x1000)
        sec_header_dict['SizeOfRawData'] = align(seg['p_filesz'] + lpad, 0x200)
        if lpad != 0:
            logger.warning(
                'ELF segment %d not aligned with page boundary. '
                'Manually padding it. This will slightly increase the executable size. '
                'Please consider rebuilding the file with the appropriate page size.', idx)
        # To be fixed in pass 2
        # sec_header_dict['PointerToRawData']
        section_dicts.append(sec_header_dict)

    optional_header_dict['SizeOfCode'] = text_size
    optional_header_dict['SizeOfInitializedData'] = data_size
    optional_header_dict['SizeOfUninitializedData'] = bss_size
    optional_header_dict['AddressOfEntryPoint'] = elf['e_entry'] - image_base
    optional_header_dict['BaseOfCode'] = text_base
    optional_header_dict['BaseOfData'] = data_base
    optional_header_dict['ImageBase'] = image_base
    optional_header_dict['SectionAlignment'] = 0x1000
    # To be fixed in pass 2
    #optional_header_dict['SizeOfImage']
    #optional_header_dict['SizeOfHeaders']

    if romspec is not None:
        # Generate rsrc data
        rsrc_emitter = BinaryBuilder()

        # /
        rsrc_root_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY.copy()
        rsrc_root_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_SIZE)

        # /BIN
        rsrc_root_dict['NumberOfNamedEntries'] += 1
        rsrc_bin_entry_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY_ENTRY.copy()
        rsrc_bin_entry_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_ENTRY_SIZE)

        rsrc_bin_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY.copy()
        rsrc_bin_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_SIZE)
        rsrc_bin_entry_dict['OffsetToData'] = IMAGE_RESOURCE_DIRECTORY_ENTRY_ISDIR | rsrc_bin_fragment.offset

        # /BIN/ROMSPC
        rsrc_bin_dict['NumberOfNamedEntries'] += 1
        rsrc_bin_romspc_entry_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY_ENTRY.copy()
        rsrc_bin_romspc_entry_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_ENTRY_SIZE)

        rsrc_bin_romspc_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY.copy()
        rsrc_bin_romspc_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_SIZE)
        rsrc_bin_romspc_entry_dict['OffsetToData'] = IMAGE_RESOURCE_DIRECTORY_ENTRY_ISDIR | rsrc_bin_romspc_fragment.offset

        # /BIN/ROMSPC/0x0401
        rsrc_bin_romspc_dict['NumberOfIdEntries'] += 1
        rsrc_bin_romspc_0404_entry_dict = EMPTY_IMAGE_RESOURCE_DIRECTORY_ENTRY.copy()
        rsrc_bin_romspc_0404_entry_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DIRECTORY_ENTRY_SIZE)

        rsrc_bin_romspc_0404_data_entry_dict = EMPTY_IMAGE_RESOURCE_DATA_ENTRY.copy()
        rsrc_bin_romspc_0404_data_entry_fragment = rsrc_emitter.append(IMAGE_RESOURCE_DATA_ENTRY_SIZE)
        rsrc_bin_romspc_0404_entry_dict['OffsetToData'] = rsrc_bin_romspc_0404_data_entry_fragment.offset

        # Set up IDs and names
        rsrc_name_bin_fragment = rsrc_emitter.append(len(RSRC_NAME_BIN))
        rsrc_name_bin_fragment.set_data(RSRC_NAME_BIN)
        rsrc_bin_entry_dict['Name'] = IMAGE_RESOURCE_DIRECTORY_ENTRY_ISNAME | rsrc_name_bin_fragment.offset

        rsrc_name_romspc_fragment = rsrc_emitter.append(len(RSRC_NAME_ROMSPC))
        rsrc_name_romspc_fragment.set_data(RSRC_NAME_ROMSPC)
        rsrc_bin_romspc_entry_dict['Name'] = IMAGE_RESOURCE_DIRECTORY_ENTRY_ISNAME | rsrc_name_romspc_fragment.offset

        _rsrc_name_nil_fragment = rsrc_emitter.append(len(RSRC_NAME_NIL))
        _rsrc_name_nil_fragment.set_data(RSRC_NAME_NIL)

        rsrc_bin_romspc_0404_entry_dict['Name'] = 0x0404

        rsrc_romspc_data_fragment = rsrc_emitter.append(len(romspec))
        rsrc_bin_romspc_0404_data_entry_dict['OffsetToData'] = rsrc_base + rsrc_romspc_data_fragment.offset
        rsrc_bin_romspc_0404_data_entry_dict['Size'] = rsrc_romspc_data_fragment.size

        # Serialize headers
        rsrc_root_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_format__,
            rsrc_root_dict,
            file_offset=rsrc_base + rsrc_root_fragment.offset,
        ).__pack__())

        rsrc_bin_entry_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_ENTRY_format__,
            rsrc_bin_entry_dict,
            file_offset=rsrc_base + rsrc_bin_entry_fragment.offset,
        ).__pack__())

        rsrc_bin_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_format__,
            rsrc_bin_dict,
            file_offset=rsrc_base + rsrc_bin_fragment.offset,
        ).__pack__())

        rsrc_bin_romspc_entry_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_ENTRY_format__,
            rsrc_bin_romspc_entry_dict,
            file_offset=rsrc_base + rsrc_bin_romspc_entry_fragment.offset,
        ).__pack__())

        rsrc_bin_romspc_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_format__,
            rsrc_bin_romspc_dict,
            file_offset=rsrc_base + rsrc_bin_romspc_fragment.offset,
        ).__pack__())

        rsrc_bin_romspc_0404_entry_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DIRECTORY_ENTRY_format__,
            rsrc_bin_romspc_0404_entry_dict,
            file_offset=rsrc_base + rsrc_bin_romspc_0404_entry_fragment.offset,
        ).__pack__())

        rsrc_bin_romspc_0404_data_entry_fragment.set_data(pefile_struct_from_dict(
            pefile.PE.__IMAGE_RESOURCE_DATA_ENTRY_format__,
            rsrc_bin_romspc_0404_data_entry_dict,
            file_offset=rsrc_base + rsrc_bin_romspc_0404_data_entry_fragment.offset,
        ).__pack__())

        rsrc_romspc_data_fragment.set_data(romspec)

        rsrc_data = (rsrc_emitter.concat(), 0, False)

    if rsrc_data is not None:
        # Calculate rsrc allocation size
        rsrc_memsize = len(rsrc_data[0])
        rsrc_size = align(rsrc_memsize, 0x200)
        rsrc_vsize = align(rsrc_memsize, 0x1000)
        logger.debug('rsrc_memsize = %s, rsrc_vsize = %s', hex(rsrc_memsize), hex(rsrc_vsize))
    else:
        rsrc_memsize = rsrc_size = rsrc_vsize = 0

    if rsrc_data is not None:
        # Append rsrc header
        rsrc_section_header = EMPTY_SECTION_HEADER.copy()
        rsrc_section_header['Name'] = b'.rsrc'
        rsrc_section_header['Characteristics'] = (
            pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA'] |
                pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ']
        )
        rsrc_section_header['Misc_VirtualSize'] = rsrc_section_header['Misc_PhysicalAddress'] = rsrc_section_header['Misc'] = rsrc_memsize
        rsrc_section_header['VirtualAddress'] = rsrc_base
        rsrc_section_header['SizeOfRawData'] = rsrc_size
        section_dicts.append(rsrc_section_header)

    reloc_base = rsrc_base + rsrc_vsize

    reloc_dicts: Dict[int, Dict[int, int]] = defaultdict(dict)
    # Generate the reloc table
    for rel_section in elf.iter_sections():
        if isinstance(rel_section, RelocationSection):
            logger.debug('Processing section %s', rel_section.name)
            # Complain on RELA
            if rel_section.is_RELA():
                raise RuntimeError('RELA relocation is not supported.')

            rel_target_sec = elf.get_section(rel_section['sh_info'])
            # Skip non-loadable sections
            if (rel_target_sec['sh_flags'] & elfconsts.SH_FLAGS.SHF_ALLOC) == 0:
                continue
            rel_target_data = rel_target_sec.data()

            if rel_section.name in ('.rel.ARM.exidx', '.rel.ARM.extab', '.rel.eh_frame'):
                # exidx and extab are handled by libgcc's unwind routine. Do nothing here.
                logger.info('Skipping exception related section.')
                continue

            for reloc in rel_section.iter_relocations():
                symtab = elf.get_section(rel_section['sh_link'])
                sym, type_ = symtab.get_symbol(reloc['r_info_sym']), reloc['r_info_type']
                sym_offset = sym['st_value']
                rel_offset = reloc['r_offset']
                rel_word_offset = rel_offset - rel_target_sec['sh_addr']
                rel_target_word = int.from_bytes(rel_target_data[rel_word_offset:rel_word_offset+4], 'little')

                if rel_word_offset < 0 or rel_word_offset >= len(rel_target_data):
                    logger.info('Ignoring out of bound symbol @ %#x', rel_word_offset)
                    continue

                if type_ in AAELF_RELOC_ABSOLUTE:
                    logger.debug('Abs reloc type=%s, value=%#010x, sym_value=%#010x, @ %#010x (%#010x in section)', ENUM_RELOC_NAME_ARM[type_], rel_target_word, sym_offset, rel_offset, rel_word_offset)
                    if sym['st_info']['bind'] == 'STB_WEAK' and sym_offset == 0:
                        logger.debug('Skipping weak symbols.')
                        continue

                    assert rel_target_word >= elf_base, 'REL symbol is not a valid address within the program.'
                    rel_offset_hi = rel_offset & 0xfffff000
                    rel_offset_lo = rel_offset & 0x00000fff
                    reloc_dicts[rel_offset_hi][rel_offset_lo] = pefile.RELOCATION_TYPE['IMAGE_REL_BASED_HIGHLOW']

                elif type_ == R_ARM_PREL31:
                    logger.info('Ignoring non-exception PREL31 (%#010x @ %#010x).', rel_target_word, rel_offset)

                elif type_ in AAELF_RELOC_RELATIVE:
                    logger.info('Ignoring relative reloc (%#010x @ %#010x).', rel_target_word, rel_offset)

                elif type_ in AAELF_RELOC_IGNORE:
                    continue
                else:
                    raise RuntimeError(f'Unhandled reloc type {ENUM_RELOC_NAME_ARM[type_]}')

    # Actually generate relocs
    generated_relocs: List[pefile.Structure] = []
    reloc_pos = 0
    for hi, los in reloc_dicts.items():
        base_reloc_struct = pefile_struct_from_dict(pefile.PE.__IMAGE_BASE_RELOCATION_format__, {'VirtualAddress': hi - image_base, 'SizeOfBlock': 0}, file_offset=reloc_pos)
        generated_relocs.append(base_reloc_struct)
        base_reloc_struct.SizeOfBlock = 8
        reloc_pos += generated_relocs[-1].sizeof()
        for lo, type_ in los.items():
            generated_relocs.append(pefile_struct_from_dict(pefile.PE.__IMAGE_BASE_RELOCATION_ENTRY_format__, {'Data': (type_ << 12 | lo)}, file_offset=reloc_pos))
            reloc_pos += generated_relocs[-1].sizeof()
            base_reloc_struct.SizeOfBlock += generated_relocs[-1].sizeof()
        if len(los) % 2 == 1:
            # pad the entry so the reloc table is always 4 byte aligned
            generated_relocs.append(pefile_struct_from_dict(pefile.PE.__IMAGE_BASE_RELOCATION_ENTRY_format__, {'Data': 0}, file_offset=reloc_pos))
            reloc_pos += generated_relocs[-1].sizeof()
            base_reloc_struct.SizeOfBlock += generated_relocs[-1].sizeof()

    relocs_data_io = io.BytesIO()
    for reloc_struct in generated_relocs:
        relocs_data_io.write(reloc_struct.__pack__())

    reloc_data = (relocs_data_io.getvalue(), 0, True)
    reloc_size = align(reloc_pos, 0x200, greedy=True)
    reloc_memsize = reloc_pos

    reloc_section_header = EMPTY_SECTION_HEADER.copy()
    reloc_section_header['Name'] = b'.reloc'
    reloc_section_header['Characteristics'] = (
        pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA'] |
            pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_DISCARDABLE'] |
            pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ']
    )
    reloc_section_header['Misc_VirtualSize'] = reloc_section_header['Misc_PhysicalAddress'] = reloc_section_header['Misc'] = reloc_memsize
    reloc_section_header['VirtualAddress'] = reloc_base
    reloc_section_header['SizeOfRawData'] = reloc_size
    section_dicts.append(reloc_section_header)

    pos = 0x0

    dos_header = pefile_struct_from_dict(pefile.PE.__IMAGE_DOS_HEADER_format__, EMPTY_DOS_HEADER, file_offset=pos)
    pos += dos_header.sizeof()

    nt_header = pefile_struct_from_dict(pefile.PE.__IMAGE_NT_HEADERS_format__, NT_HEADER, file_offset=pos)
    pos += nt_header.sizeof()

    file_header_dict = EMPTY_FILE_HEADER.copy()
    # Number of sections will be filled in later

    if not args.deterministic:
        file_header_dict['TimeDateStamp'] = int(datetime.datetime.now().timestamp())

    file_header_dict['Characteristics'] = (
        pefile.IMAGE_CHARACTERISTICS['IMAGE_FILE_EXECUTABLE_IMAGE'] |
            pefile.IMAGE_CHARACTERISTICS['IMAGE_FILE_32BIT_MACHINE'] |
            pefile.IMAGE_CHARACTERISTICS['IMAGE_FILE_LINE_NUMS_STRIPPED'] |
            pefile.IMAGE_CHARACTERISTICS['IMAGE_FILE_LOCAL_SYMS_STRIPPED'] |
            pefile.IMAGE_CHARACTERISTICS['IMAGE_FILE_DEBUG_STRIPPED']
    )
    file_header_dict['NumberOfSections'] = len(section_dicts)
    file_header = pefile_struct_from_dict(pefile.PE.__IMAGE_FILE_HEADER_format__, file_header_dict, file_offset=pos)
    pos += file_header.sizeof()

    optional_header = pefile_struct_from_dict(pefile.PE.__IMAGE_OPTIONAL_HEADER_format__, optional_header_dict, file_offset=pos)
    pos += optional_header.sizeof()

    directories: List[pefile.Structure] = []
    for _ in range(pefile.IMAGE_NUMBEROF_DIRECTORY_ENTRIES):
        dir_ = pefile_struct_from_dict(pefile.PE.__IMAGE_DATA_DIRECTORY_format__, {'VirtualAddress': 0, 'Size': 0}, file_offset=pos)
        pos += dir_.sizeof()
        directories.append(dir_)

    if rsrc_data is not None:
        directories[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']].VirtualAddress = rsrc_base
        directories[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_RESOURCE']].Size = rsrc_memsize
    directories[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_BASERELOC']].VirtualAddress = reloc_base
    directories[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_BASERELOC']].Size = reloc_memsize

    # Set optional header size
    file_header.SizeOfOptionalHeader = optional_header.sizeof() + sum(dir_.sizeof() for dir_ in directories)

    sections: List[pefile.Structure] = []

    for sec in section_dicts:
        pestruct = pefile_struct_from_dict(pefile.PE.__IMAGE_SECTION_HEADER_format__, sec, file_offset=pos)
        sections.append(pestruct)
        pos += pestruct.sizeof()

    pos = align(pos, 0x200)

    # Pass 2: fixing field values that were not yet available during pass 1 and prepare the actual section data
    header_size = pos
    optional_header.SizeOfHeaders = header_size

    for sec_struct in sections:
        sec_struct.PointerToRawData = pos
        pos += sec_struct.SizeOfRawData # pylint:disable=invalid-name

    image_size = pos
    optional_header.SizeOfImage = align(reloc_base + reloc_memsize, 0x1000)

    logger.debug("File Header:")
    logger.debug(file_header)
    logger.debug('Optional Header:')
    logger.debug(optional_header)
    logger.debug('Directory Header:')
    for dir_ in directories:
        logger.debug(dir_)
    logger.debug('Section Header:')
    for sec in sections:
        logger.debug(sec)

    all_headers = (
        dos_header,
        nt_header,
        file_header,
        optional_header,
        *directories,
        *sections,
    )

    for hdr in all_headers:
        pe_file.write(hdr.__pack__())

    pe_file.write(generate_padding(pe_file.tell(), 0x200))

    for data in (text_data, rdata_data, data_data, rsrc_data, reloc_data):
        if data is not None:
            greedy = data[2]
            lpad = data[1]
            if lpad != 0:
                pe_file.write(b'\x00' * lpad)
            pe_file.write(data[0])
            pe_file.write(generate_padding(pe_file.tell(), 0x200, greedy=greedy))
    actual_image_size = len(pe_file.getvalue())
    assert actual_image_size == image_size, f'Inconsistent generated image size vs calculated (expecting {image_size:#x}, got {actual_image_size:#x}).'

    # Pass 3: using pefile for some fixing and linting
    pefile_obj = pefile.PE(data=pe_file.getvalue())
    pefile_obj.OPTIONAL_HEADER.CheckSum = pefile_obj.generate_checksum()

    for offset, val in patches.items():
        rva = offset - image_base
        logger.debug('Postprocessing: patch %d bytes at %#010x (RVA %#010x)', len(val), offset, rva)
        pefile_obj.set_bytes_at_rva(rva, val)

    pe_file.truncate(0)
    pe_file.seek(0)

    logger.debug('pefile objdump:\n%s', pefile_obj.dump_info())
    for pefile_warning in pefile_obj.get_warnings():
        if segment_load_file_size < 0x8000 and PEFILE_WARNING_TOO_MANY_0.match(pefile_warning) is not None:
            logger.debug('suppressed pefile warning: %s', pefile_warning)
            continue
        logger.info('pefile warning: %s', pefile_warning)
    pe_file.write(pefile_obj.write())

def main():
    _, args = parse_args()

    logging.basicConfig()
    if args.log_level is not None:
        logger.setLevel(args.log_level)

    if args.output is None:
        output_path = f'{os.path.splitext(args.elf)[0]}{os.path.extsep}exe'
    else:
        output_path = args.output

    if args.romspec_file is not None:
        if os.stat(args.romspec_file).st_size > BESTAPE_MAX_ROMSPEC_SIZE:
            raise RuntimeError('Refusing to link ROM spec file of size greater than 32KiB.')
        with open(args.romspec_file, 'rb') as romspec_file:
            romspec = romspec_file.read()
    else:
        romspec = None

    with open(args.elf, 'rb') as elf_file:
        pe_file = io.BytesIO()
        convert(elf_file, pe_file, romspec, args)
    with open(output_path, 'wb') as actual_pe_file:
        pe_file.seek(0)
        shutil.copyfileobj(pe_file, actual_pe_file)

if __name__ == '__main__':
    main()



from typing import cast, TYPE_CHECKING
if TYPE_CHECKING:
    from construct import Context

import dataclasses
from enum import IntEnum

from construct import (
    Bytes,
    Default,
    Int32ul,
    this
)
from construct_typed import DataclassMixin, DataclassStruct, EnumBase, TEnum, csfield


class BestaDfuSbcOpcode(IntEnum):
    SET_CONFIG = 0x88
    GET_CONFIG = 0x89


class BestaDfuCommand(EnumBase):
    CMD_PING_ARG = 0xbead
    CMD_ERASE_AND_SCAN = 0xbec2
    CMD_PROBE_REGION = 0xbec5
    CMD_UPLOAD_BOOTLOADER = 0xbec6
    CMD_COMMIT_BOOTLOADER = 0xbec7
    CMD_PING = 0xbecd
    CMD_SET_PROGRESS = 0xbedc
    ACK = 0x80000000


CsBestaDfuCommand = TEnum(Int32ul, BestaDfuCommand)


@dataclasses.dataclass
class BestaDfuConfigPacket(DataclassMixin):
    command: int = csfield(CsBestaDfuCommand)
    parameter: int = csfield(Int32ul)
    payload: bytes = csfield(Default(Bytes(256), b'\x00' * 256))


CsBestaDfuConfigPacket = DataclassStruct(BestaDfuConfigPacket)



from typing import TextIO, BinaryIO, Any, Self, Optional, Literal

import click
import dataclasses
import pathlib
import sys

from ..common.utils import simple_checksum


def relative_to_root(path: pathlib.PureWindowsPath) -> pathlib.PureWindowsPath:
    return path.relative_to(path.anchor)


def search_file(root: pathlib.Path, rel_path: str | pathlib.Path):
    if isinstance(rel_path, pathlib.Path) or isinstance(rel_path, pathlib.PurePath):
        rel_path = str(rel_path.as_posix())
    files = tuple(root.glob(rel_path, case_sensitive=False))
    if len(files) == 0:
        return root / rel_path
    return files[0]


class ParserError(RuntimeError):
    pass


@dataclasses.dataclass
class TitleChecksumEntry:
    path: pathlib.PureWindowsPath
    checksum: int


@dataclasses.dataclass
class TitleIndex:
    version: str
    overall_checksum: int
    entries: list[TitleChecksumEntry]

    @classmethod
    def from_scratch(cls) -> Self:
        return cls('V0.00', 0, [])

    @classmethod
    def from_file(cls, file: TextIO) -> Self:
        version = file.readline()
        overall_checksum = int(file.readline(), 16)
        num_files = int(file.readline())
        entries: list[TitleChecksumEntry] = []
        for i in range(num_files):
            path = file.readline()
            checksum = file.readline()
            entries.append(TitleChecksumEntry(pathlib.PureWindowsPath(path.rstrip()), int(checksum, 16)))
        return cls(version, overall_checksum, entries)

    def dump(self) -> str:
        result: list[str] = [
            self.version,
            f'{self.overall_checksum:#06x}',
            f'{len(self.entries):>6}'
        ]

        for entry in self.entries:
            result.append(entry.path)
            result.append(f'{entry.checksum:#06x}')
        return '\n'.join(result)

    def to_file(self, file: TextIO) -> None:
        file.write(self.dump())

    def validate(self) -> bool:
        overall_checksum = sum(entry.checksum for entry in self.entries) & 0xffff
        return overall_checksum == self.overall_checksum

    def fix_checksum(self) -> None:
        self.overall_checksum = sum(entry.checksum for entry in self.entries) & 0xffff


@click.group()
def app():
    pass


@app.command('init', help='Initialize an title index file.')
@click.argument('path', type=click.Path(exists=True))
@click.option('-f', '--force', is_flag=True, help='Overwrite any existing title index file.')
def do_init(path: str, force: bool):
    root = pathlib.Path(path)
    index_file = search_file(root, 'SYSCHECK.$$$')

    if index_file.is_file():
        if not force:
            click.echo('Refusing to create a new index file as it already exists.')
            sys.exit(1)
    elif index_file.exists():
        click.echo('Index file does not seem to be a file. Aborting.')
        sys.exit(1)

    with index_file.open('w', newline='\r\n', encoding='ascii') as f:
        TitleIndex.from_scratch().to_file(f)


@app.command('validate', help='Validate a user data root.')
@click.argument('path', type=click.Path(exists=True))
@click.option('-m', '--metadata-only', is_flag=True, help='Only check index consistency.')
@click.option('-e', '--encoding', default='cp936', help='Encoding type to use (default is cp936).')
def do_validate(path: str, metadata_only: bool, encoding: str):
    root = pathlib.Path(path)
    index_file = search_file(root, 'SYSCHECK.$$$')
    with index_file.open('r', encoding=encoding) as f:
        title_index = TitleIndex.from_file(f)

    if not title_index.validate():
        click.echo('Overall checksum inconsistent with the actual sum of checksum of entries.')
        sys.exit(1)

    if metadata_only:
        return
    
    failed_io = 0
    failed_checksum = 0

    for entry in title_index.entries:
        actual_path = search_file(root, relative_to_root(entry.path))
        try:
            with actual_path.open('rb') as f:
                checksum = simple_checksum(f)
                if checksum != entry.checksum:
                    click.echo(f'{str(actual_path)}: FAILED')
                    failed_checksum += 1
                else:
                    click.echo(f'{str(actual_path)}: OK')
        except OSError as err:
            click.echo(f'Validation error: {repr(str(actual_path))}: {err.strerror}')
            click.echo(f'{str(actual_path)}: FAILED open or read')
            failed_io += 1

    if failed_io > 0:
        click.echo(f'WARNING: {failed_io} file(s) could not be read.')
    if failed_checksum > 0:
        click.echo(f'WARNING: {failed_checksum} computed checksum did NOT match.')


def main():
    app()



from typing import Final, TypedDict, NotRequired, Sequence, cast, Mapping

import io

from .formats import RomExtMetadataHeader, CsRomSpecType, CsRomExtMetadataHeader, CsRomLocalizedTitle, \
    CsRomLocalizedTitleEntry, CsRomFallbackTitle, RomLocalizedTitle, RomLocale, RomFallbackTitle, RomSpecType, \
    MagicType, RomType, RomSpecTimestamp, RomLocalizedTitleEntry
from ..common.formats import ChecksumValue
from ..common.utils import simple_checksum, Fragment, BinaryBuilder

LOCALE_MAPPING: Final[dict[str, RomLocale]] = {
    'zh_CN': RomLocale.ZH_CN,
    'zh_TW': RomLocale.ZH_TW,
    'ja': RomLocale.JA_JP,
    'ko': RomLocale.KO_KR,
    'th': RomLocale.TH_TH,
    'en': RomLocale.EN_US,
}

SpecTomlRomSection = TypedDict('SpecTomlRomSection', {
    'title': str,
    'short-title': NotRequired[str],
    'title-alt': NotRequired[str],
    'short-title-alt': NotRequired[str],
    'category': str | int,
    'subtype': NotRequired[int],
    'version': str,
    'copyright': str,
    'sdk-id': NotRequired[str],
    'default-locale': NotRequired[str | int],
    'icon': NotRequired[str],
})

SpecTomlMetadataElement = TypedDict('SpecTomlMetadataElement', {
    'title': str,
    'short-title': NotRequired[str],
})


class SpecToml(TypedDict):
    rom: SpecTomlRomSection
    metadata: NotRequired[Mapping[str, SpecTomlMetadataElement]]


def build_embeddable_from_spec_file(spec_dict_in: dict):
    str_table: dict[str, Fragment] = {}

    def add_to_strtab(str_: str) -> Fragment:
        nonlocal str_table
        if str_ not in str_table:
            encoded = str_.encode('utf-16le')
            str_table_entry = builder.append(len(encoded) + 2)
            str_table_entry.set_data(encoded + b'\x00\x00')
            str_table[str_] = str_table_entry
            return str_table_entry
        else:
            return str_table[str_]

    spec_dict = cast(SpecToml, spec_dict_in)
    spec_dict_rom = spec_dict['rom']
    spec_dict_metadata = spec_dict.get('metadata', [])

    build_date = RomSpecTimestamp.now()

    builder = BinaryBuilder()

    # Forward declaration
    header_alloc = builder.append(CsRomSpecType.sizeof())
    ext_metadata_header_alloc = builder.append(CsRomExtMetadataHeader.sizeof())
    localized_title_alloc = builder.append(CsRomLocalizedTitle.sizeof())
    localized_title_elements_alloc = [
        builder.append(CsRomLocalizedTitleEntry.sizeof()) for _ in spec_dict_metadata
    ]
    fallback_title_alloc = builder.append(CsRomFallbackTitle.sizeof())

    ext_metadata_header = RomExtMetadataHeader(
        num_localized_title_entries=len(spec_dict_metadata),
    )
    ext_metadata_header_alloc.set_data(CsRomExtMetadataHeader.build(ext_metadata_header))

    localized_title = RomLocalizedTitle(
        num_entries=len(spec_dict_metadata),
        locale=RomLocale.UNSET.value,
    )
    localized_title_alloc.set_data(CsRomLocalizedTitle.build(localized_title))

    for element, (locale_str, metadata) in zip(localized_title_elements_alloc, spec_dict_metadata.items()):
        title_fragment = add_to_strtab(metadata['title'])
        short_title_fragment = add_to_strtab(metadata['short-title'])

        encoded_element_index = RomLocalizedTitleEntry(
            locale=LOCALE_MAPPING[locale_str],
            title_offset=title_fragment.offset,
            short_title_offset=short_title_fragment.offset,
        )
        element.set_data(CsRomLocalizedTitleEntry.build(encoded_element_index))

    fallback_title_fragment = add_to_strtab(spec_dict_rom['title'])

    if 'title-alt' in spec_dict_rom:
        fallback_title_alt_fragment = add_to_strtab(spec_dict_rom['title-alt'])
    else:
        fallback_title_alt_fragment = fallback_title_fragment

    if 'short-title' in spec_dict_rom:
        fallback_short_title_fragment = add_to_strtab(spec_dict_rom['short-title'])
    else:
        fallback_short_title_fragment = None

    if 'short-title-alt' in spec_dict_rom:
        fallback_short_title_alt_fragment = add_to_strtab(spec_dict_rom['short-title-alt'])
    else:
        fallback_short_title_alt_fragment = None

    fallback_title_index = RomFallbackTitle(
        locale=RomLocale.UNSET.value,
        title_offset=fallback_title_fragment.offset,
        short_title_offset=(
            fallback_short_title_fragment.offset if fallback_short_title_fragment is not None else 0xffffffff
        ),
        chinese_title_offset=fallback_title_alt_fragment.offset,
        chinese_short_title_offset=(
            fallback_short_title_alt_fragment.offset if fallback_short_title_alt_fragment is not None else 0xffffffff
        ),
    )

    fallback_title_alloc.set_data(CsRomFallbackTitle.build(fallback_title_index))

    copyright_offset = add_to_strtab(spec_dict_rom['copyright']).offset
    version_offset = add_to_strtab(spec_dict_rom['version']).offset
    sdk_id_offset = add_to_strtab(spec_dict_rom.get('sdk-id', 'besta_tools.romtool')).offset

    checksum_value = ChecksumValue(0)

    header = RomSpecType(
        magic=MagicType.APPLET,
        type_=(
            RomType[spec_dict_rom['category']].value
            if isinstance(spec_dict_rom['category'], str)
            else int(spec_dict_rom['category'])
        ),
        checksum=checksum_value,
        default_locale=RomLocale.FORCE_UTF16.value | LOCALE_MAPPING[spec_dict_rom['default-locale']].value,
        sections_offset=0,
        build_timestamp=build_date,
        rom_size=builder.sizeof(),
        entry_point=0,
        fallback_title_offset=fallback_title_alloc.offset,
        copyright_offset=copyright_offset,
        icon_offset=0xffffffff,
        version_offset=version_offset,
        data_offset=builder.sizeof(),
        sdk_id_offset=sdk_id_offset,
        ext_metadata_offset=ext_metadata_header_alloc.offset,
        localized_title_offset=localized_title_alloc.offset,
        subtype=spec_dict_rom.get('subtype', 0),
    )

    header_alloc.set_data(CsRomSpecType.build(header))

    # Fix checksum
    checksum = simple_checksum(io.BytesIO(builder.concat()))
    checksum_value.checksum = checksum

    return builder.concat()



from typing import BinaryIO, Self, cast, TYPE_CHECKING
if TYPE_CHECKING:
    from construct import Context

from dataclasses import dataclass, field
from pathlib import Path

import yaml

from construct import (
    Array,
    Computed,
    Const,
    Check,
    IfThenElse,
    Int32ul,
    Int64ul,
    PaddedString,
    Padding,
    Rebuild,
    this,
    len_
)
from construct_typed import DataclassMixin, DataclassStruct, csfield
from marshmallow_dataclass import dataclass as mm_dataclass
from marshmallow_dataclass import class_schema as mm_class_schema

from ..common.formats import CsChecksumValue, ChecksumValue
from ..common.utils import simple_checksum


IMAGE_INDEX_V2_MAGIC = b'\xaa\x55\xaa\x55'


def guess_block_size_image_v2(f: BinaryIO, search_limit: int = 0x100000, step_size: int = 16) -> int:
    for offset in range(0, search_limit, step_size):
        f.seek(offset)
        if f.read(step_size)[:4] == IMAGE_INDEX_V2_MAGIC:
            return offset
    raise RuntimeError('Cannot determine block size.')


def _inv_u16_per_byte(ctx: 'Context') -> int:
    value = cast(int, ctx.checksum)
    lo = value & 0xff
    hi = (value >> 8) & 0xff
    return (((0x100 - hi) & 0xff) << 8) | ((0x100 - lo) & 0xff)


@dataclass
class ImageMetadataV2(DataclassMixin):
    image_name: str = csfield(PaddedString(16, 'ascii'))
    image_version: str = csfield(PaddedString(16, 'ascii'))
    os_version: str = csfield(PaddedString(16, 'ascii'))
    content_size: int = csfield(Int64ul)
    data_size: int = csfield(Int64ul)
    checksum_block_size: int = csfield(Int32ul)
    unk_0x44: int = csfield(Int32ul)
    _padding_0x48: int = csfield(Padding(24))
    has_checksum: bool = csfield(Computed(this.checksum_block_size != 0))
    checksums: list[ChecksumValue] = csfield(IfThenElse(
        this.has_checksum,
        Array(this.data_size // this.checksum_block_size, CsChecksumValue),
        Array(0, CsChecksumValue),
    ))


CsImageMetadataV2 = DataclassStruct(ImageMetadataV2)


@dataclass
class ImageIndexEntryV2(DataclassMixin):
    offset: int = csfield(Int64ul)
    size: int = csfield(Int64ul)
    _integrity: None = csfield(Check(this.offset != (1 << 64) - 1 and this.size != (1 << 64) - 1))


CsImageIndexEntryV2 = DataclassStruct(ImageIndexEntryV2)


IMAGE_INDEX_V2_VERSIONS: set[int] = {
    0x20090828,
    0x20081202,
}


@dataclass
class ImageIndexV2(DataclassMixin):
    magic: bytes = csfield(Const(IMAGE_INDEX_V2_MAGIC))
    header_size: int = csfield(Const(0x40, Int32ul))
    format_version: int = csfield(Int32ul)
    _integrity_supported_format_version: None = csfield(Check(
        lambda ctx: ctx.format_version in IMAGE_INDEX_V2_VERSIONS
    ))
    nentries: int = csfield(Rebuild(Int32ul, len_(this.entries)))
    _padding_0x10: None = csfield(Padding(0x30))
    entries: list[ImageIndexEntryV2] = csfield(Array(this.nentries, CsImageIndexEntryV2))


CsImageIndexV2 = DataclassStruct(ImageIndexV2)


@mm_dataclass
class ImageManifestSection:
    path: str
    align: int = field(default=1, metadata={'required': False})


@mm_dataclass
class ImageManifest:
    header_format_version: int
    index_format_version: int
    name: str
    version_string: str
    block_size: int
    checksum_block_size: int
    sections: list[ImageManifestSection]


MmImageManifest = mm_class_schema(ImageManifest)


@dataclass
class ImageFileV2:
    path: Path
    metadata: ImageMetadataV2 | None = field(default=None)
    index: ImageIndexV2 | None = field(default=None)
    manifest: ImageManifest | None = field(default=None)

    def __post_init__(self):
        if self.manifest is None:
            self._build_manifest()

    def _build_manifest(self):
        with self.path.open('rb') as f:
            guessed_block_size = guess_block_size_image_v2(f)

        sections: list[ImageManifestSection] = [
            ImageManifestSection(f'sections/{i:08x}.bin') for i in range(self.index.nentries)
        ]

        self.manifest = ImageManifest(
            2,
            self.index.format_version,
            self.metadata.image_name,
            self.metadata.image_version,
            guessed_block_size,
            self.metadata.checksum_block_size,
            sections
        )

    @classmethod
    def load(cls, image_path: str | Path) -> Self:
        image_path = Path(image_path)
        with image_path.open('rb') as f:
            block1_offset = guess_block_size_image_v2(f)
            f.seek(0)
            metadata = CsImageMetadataV2.parse_stream(f)
            f.seek(block1_offset)
            index = CsImageIndexV2.parse_stream(f)

        return cls(image_path, metadata=metadata, index=index)

    def build(self, output: str | Path) -> None:
        ...

    def extract(self, output: str | Path) -> None:
        output = Path(output)
        output.mkdir(exist_ok=True)
        with (output / 'manifest.yaml').open('w') as manifest_file:
            yaml.safe_dump(MmImageManifest().dump(self.manifest), manifest_file, sort_keys=False)
        with self.path.open('rb') as image_file:
            for section, entry in zip(self.manifest.sections, self.index.entries):
                section_file_path = output / section.path
                section_file_path.parent.mkdir(exist_ok=True)
                with section_file_path.open('wb') as section_file:
                    image_file.seek(entry.offset)
                    bytes_left: int = entry.size
                    while bytes_left > 0:
                        bytes_to_read = min(4096, bytes_left)
                        section_file.write(image_file.read(bytes_to_read))
                        bytes_left -= bytes_to_read

    @classmethod
    def from_manifest(cls, manifest_path: str | Path) -> Self:
        ...

    def verify(self) -> list[bool]:
        results: list[bool] = []
        with self.path.open('rb') as f:
            for offset, checksum in zip(
                    range(0, self.metadata.data_size, self.metadata.checksum_block_size),
                    self.metadata.checksums
            ):
                f.seek(offset)
                checksum_actual = simple_checksum(f, self.metadata.checksum_block_size)
                results.append(checksum_actual == checksum.checksum)
        return results
